// Custom Alphabet
// by forkinsocket

// Click on the image to give it focus and select a letter on the
// keyboard to transform the line into that letter.

// Updated 10 December 2002
// Created 1 November 2002

// Array of balls. A ball is a vertex (control point) on a spline curve.
Ball balls[] = new Ball[5]; 
      
// Lookup table containing the coordinates of each vertex of each letterform. 
// Used to set target values of vertices when keys are pressed.
float lookup[] = {
      // the format for each line is: x0, y0, x1, y1, ... , x4, y4
      88, 217.95, 148, 65, 205, 218.95, 187, 137.95, 70, 152.95, // a 
      80, 179.95, 125, 118.95, 213, 178.95, 121, 245.95, 121, 10, // b
      204, 69.95, 76, 138.95, 204, 209.95, 204, 209.95, 204, 209.95, // c    
      85, 207.95, 63, 74.95, 206, 140.95, 85, 207.95, 85, 207.95, // d
      75, 142.95, 198, 131.95, 201, 68, 76, 132.95, 219, 215.95, // e
      209, 66, 71, 144.95, 73, 240.95, 56, 146.95, 208, 142.95, // f
      205, 69.95, 79, 135.95, 210, 213.95, 222, 145.95, 177, 138.95, // g
      118, 40, 106, 237.95, 105, 168.95, 181, 131.95, 209, 238.95, // h
      160, 56, 152, 119.95, 157, 221.95, 150, 263.95, 150, 263.95, // i
      163, 58, 169, 155.95, 164, 212.95, 117, 249.95, 93, 252.95, // j
      128, 51, 105, 244.95, 182, 124.95, 117, 169.95, 191, 236.95, // k
      85, 253.95, 148, 33, 110, 115.95, 167, 246.95, 187, 245.95, // l
      75, 201.95, 82, 50, 146, 201.95, 211, 50, 218, 201.95, // m
      84, 206.95, 91, 51, 198, 206.95, 209, 61, 209, 61, // n
      150, 190.95, 70, 119.95, 151, 46, 220, 122.95, 132, 184.95, // o
      84, 224.95, 107, 63, 211, 59, 220, 139.95, 87, 126.95, // p
      236, 222.95, 233, 85.95, 155, 41, 87, 103.95, 235, 92.95, // q
      77, 43, 77, 220.95, 73, 123.95, 162, 56, 223, 70.95, // r
      209, 54, 73, 85.95, 211, 184.95, 69, 216.95, 69, 216.95, // s
      170, 66, 151, 144.95, 153, 240.95, 136, 146.95, 205, 132.95, // t
      78, 79.95, 93, 207.95, 190, 207.95, 210, 69.95, 208, 227.95, // u
      113, 79.95, 155, 233.95, 190, 99.95, 190, 99.95, 190, 99.95, // v
      71, 67, 121, 203.95, 150, 67, 182, 203.95, 225, 67, // w
      50, 69.95, 201, 201.95, 200, 69.95, 68, 181.95, 68, 181.95, // x
      95, 79.95, 130, 148.95, 164, 143.95, 187, 84.95, 170, 238.95, // y
      99, 74.95, 99, 74.95, 210, 88.95, 86, 174.95, 220, 190.95 // z
      
};

void setup() 
{
  size(200, 200);
  stroke(175, 202, 205);
  rectMode(CENTER_DIAMETER);

  // Initialise the array of vertices.
  for (int i = 0; i < balls.length; i++) {
    balls[i] = new Ball();
  }
  
  // Values in lookup were designed for a 300x300 area. Scale appropriately.
  for (int i = 0; i < lookup.length; i++) {
    // Turn into a value between 0 and 1.
    lookup[i] /= 300;
    // Scale x values according to width.
    if (i%2 == 0) {
      lookup[i] *= width;
    // Scale y values according to height.
    } else {
      lookup[i] *= height;
    }
  }
}

void loop() 
{
  background(100, 151, 159);
  
  // Update the positions of each vertex
  for (int i = 0; i < balls.length; i++) {
    balls[i].move();
  }
  
  // Draw the curvy letterform. Note: when dealing with catmull rom spline curves,
  // the first and last vertices need to be added to the curve twice.
  beginShape(LINE_STRIP);
  
  // The first vertex on the curve
  curveVertex(balls[0].x, balls[0].y);
  
  // All vertices
  for (int i = 0; i < balls.length; i++) {
    curveVertex(balls[i].x, balls[i].y);
  }
  
  // The last vertex on the curve
  curveVertex(balls[balls.length - 1].x, balls[balls.length - 1].y);
  endShape();
}

void mousePressed() 
{
  // When mouse is pressed, create a random form.
  for (int i = 0; i < balls.length; i++) {
  
      // Middle vertex approaches mouse
    if (i == 2) {
      balls[i].setTarget(mouseX, mouseY);
    }
    
    // Left vertices stay on left
    else if (i == 0 || i == 1) {
      balls[i].setTarget(random(width/2), random(height));
    }
    
    // Right vertices stay on right
    else {
      balls[i].setTarget(width/2 + random(width/2), random(height));
    }
  }
}

void keyReleased() 
{
  // When a key is pressed, morph the current curve into a corresponding letterform.
  // Remember that "key" is an integer value that P5 gives us for free. It contains
  // the ascii value of the last key pressed.
  resetLetterForm(key);
}

void resetLetterForm(int letter) 
{
  // =================================================================
  // Note that:
  // 1. Uppercase letters have ascii values ranging from 65 to 90
  // 2. Lowercase letters have ascii values ranging from 97 to 122
  // 3. lookup[] contains 26 letter forms 
  // We need to map ascii values to lookup indices.
  // ==================================================================    
  
  // Turn uppercase letters into lowercase ones.
  if (letter >= 65 && letter <=90) {
    letter += 32;
  }
  
  // Normalise letter so that we can access our lookup array.
  int cur = letter - 97;
  // Cur is now a number between 0 and 25.
  // Check to make sure. (Remember that there are ten values in lookup for each letter)
  if (cur >= 0 && cur < lookup.length/10) { 
    // Set new targets for each vertex, according to lookup.
    for (int i = 0; i < 5; i++) {
      balls[i].setTarget(lookup[cur*10 + 2*i], lookup[cur*10 + 2*i + 1]);
    }
  }  
}

// A ball is a container for a set of coordinates x and y, and target coordinates
// that the ball can be told to approach. This is a common data structure used in
// the trendy flash experiments these days.

class Ball 
{
  float x, y, tx, ty;  // Coords and target coords.
  boolean active;      // Is this ball active? if so, then animate.

  Ball() {
    // When a new ball is constructed, give it random coords and targets.
    x = random(width);
    y = random(height);
    tx = random(width);
    ty = random(height);
    // Animate this ball.
    active = true;
  }

  // Tell the ball to approach a new set of coords.
  void setTarget(float xin, float yin) {
    tx = xin;
    ty = yin;
    active = true;
  }

  // Move the ball if it is active.
  void move() {
    if (this.active) {
      // Calculate the horizontal and vertical distance from target.
      float xdif, ydif;
      xdif = tx - x;
      ydif = ty - y;
      // Update the coordinates incrementally. as the ball gets closer
      // to its target, the values xdif/16 and ydif/16 become smaller,
      // causing it to slow down.
      // (The value '16' is completely arbitrary. Play with different
      // numbers and see what happens.)
      x += xdif/16;
      y += ydif/16;
      // When the ball has reached a reasonably close proximity of its
      // target, tell it to snap into position and stop.
      if (abs(xdif)<0.5 && abs(ydif)<0.5) {
        x = this.tx;
        y = this.ty;
        active = false;
      }
    }// End if active
  } // End move method
} // End class


